/*
 * source.c
 *
 * Copyright (c) 1996-1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "port_before.h"

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#ifdef HAVE_STRING_H
#include <string.h>
#endif

#include <X11/IntrinsicP.h>

#include "port_after.h"

#include "ChimeraP.h"
#include "mime.h"

typedef enum
{
  DataStateWaiting,
  DataStateEnded,
  DataStateLoading,
  DataStateUnloaded
} DataState;

typedef enum
{
  SourceStateBlocked,
  SourceStateWaiting,
  SourceStateEnded,
  SourceStateLame,
  SourceStateLoading
} SourceState;

struct ChimeraSinkP
{
  MemPool           mp;
  DataState         dstate;
  ChimeraSource     ws;
  ChimeraSinkHooks  dhooks;
  void              *closure;
  ChimeraTask       wt;
  bool              locked;
  bool              destroy;
  bool              stop_callbacks;
};

struct ChimeraSourceP
{
  MemPool           mp;
  ChimeraResources  cres;
  ChimeraSourceHooks shooks;
  void              *closure;
  GList		    datas;

  /* only used if the source is blocked.  do not free. */
  ChimeraRequest    *wr;

  /* Various status goodies */
  SourceState       sstate;             /* transfer state */
  bool              cachable;           /* can this be cached? */
  bool              reusable;           /* can this be reused? */
};

/*
 * Private functions
 */
static ChimeraSource SourceCreate _ArgProto((ChimeraResources,
					     ChimeraRequest *));
static void SourceDestroy _ArgProto((ChimeraResources, ChimeraSource));
static void SourceEndTask _ArgProto((void *));
static void SourceCreateTask _ArgProto((void *));
static void SourceStopTask _ArgProto((void *));
static char *GetSourceInfo _ArgProto((ChimeraSource, char *));
static void StartBlockedSource _ArgProto((ChimeraResources));
static int StartSource _ArgProto((ChimeraResources, ChimeraSource));
static ChimeraSource TryMemoryCache _ArgProto((ChimeraResources,
					       ChimeraRequest *));

/*
 * SourceDestroy
 */
static void
SourceDestroy(cres, ws)
ChimeraResources cres;
ChimeraSource ws;
{
  ChimeraSink c;

  if (cres->printLoadMessages)
  {
    fprintf (stderr, "Source Destroy %s\n", ws->wr->url);
  }

  for (c = (ChimeraSink)GListGetHead(ws->datas); c != NULL;
       c = (ChimeraSink)GListGetNext(ws->datas))
  {
    myassert(c->dstate == DataStateUnloaded, "Data not unloaded");

    if (c->wt != NULL) TaskRemove(ws->cres, c->wt);
    MPDestroy(c->mp);    
  }

  if (ws->closure != NULL) (ws->shooks.destroy)(ws->closure);
  GListRemoveItem(cres->sources, ws);
  MPDestroy(ws->wr->mp);
  MPDestroy(ws->mp);

  return;
}

/*
 * SourceCreate
 *
 * This function gobbles up 'wr'.
 */
static ChimeraSource
SourceCreate(cres, wr)
ChimeraResources cres;
ChimeraRequest *wr;
{
  ChimeraSource c;
  int count;
  GList list;
  ChimeraSource ws;
  MemPool mp;

  mp = MPCreate();
  ws = (ChimeraSource)MPCGet(mp, sizeof(struct ChimeraSourceP));
  ws->mp = mp;
  ws->cres = cres;
  ws->cachable = false;
  ws->reusable = false;
  ws->sstate = SourceStateWaiting;
  ws->datas = GListCreateX(mp);
  ws->wr = wr;

  count = 0;
  list = cres->sources;
  for (c = (ChimeraSource)GListGetHead(list); c != NULL;
       c = (ChimeraSource)GListGetNext(list))
  {
    if (c->sstate == SourceStateLoading || c->sstate == SourceStateWaiting)
    {
      count++;
      if (count == cres->maxiocnt)
      {
	ws->sstate = SourceStateBlocked;
	GListAddHead(cres->sources, ws);
	if (cres->printLoadMessages)
	{
	  fprintf (stderr, "Source Block %s\n", wr->url);
	}
	return(ws);
      }
    }
  }

  GListAddHead(cres->sources, ws);

  if (StartSource(cres, ws) == -1) return(NULL);

  return(ws);
}

/*
 * StartSource
 */
static int
StartSource(cres, ws)
ChimeraResources cres;
ChimeraSource ws;
{
  ChimeraSourceHooks *hooks;
  ChimeraRequest *wr = ws->wr;

  /*
   * Look in the cache if the reload flag isn't set.  Copy the cache
   * source hooks to be used later.  Set the reusable flag so that the
   * memory can be used for later loads.
   */
  if (!wr->reload && wr->input_data == NULL && cres->cc != NULL)
  {
    hooks = CacheGetHooks(cres->cc);
    if (hooks != NULL)
    {
      ws->closure = (hooks->init)(ws, wr, hooks->class_closure);
      if (ws->closure != NULL)
      {
	ws->reusable = true;
	memcpy(&(ws->shooks), hooks, sizeof(ws->shooks));
      }
    }
  }

  /*
   * Cache could not be attempted or the cache grab failed.  Try the
   * real handler.
   */
  if (ws->closure == NULL)
  {
    ws->closure = (wr->hooks.init)(ws, wr, wr->hooks.class_closure);
    memcpy(&(ws->shooks), &(wr->hooks), sizeof(ws->shooks));    
  }

  /*
   * Couldn't start a source.  Just give up and return nothing.
   */
  if (ws->closure == NULL)
  {
    GListRemoveItem(cres->sources, ws);
    SourceDestroy(cres, ws);

    if (cres->printLoadMessages)
    {
      fprintf (stderr, "Source StartFailed %s\n", wr->url);
    }

    return(-1);
  }

  ws->sstate = SourceStateWaiting;

  if (cres->printLoadMessages)
  {
    fprintf (stderr, "Source StartSuccess %s\n", wr->url);
  }

  return(0);
}

/*
 * StartBlockedSource
 */
static void
StartBlockedSource(cres)
ChimeraResources cres;
{
  ChimeraSource c;

  for (c = (ChimeraSource)GListGetHead(cres->sources); c != NULL; )
  {
    if (c->sstate == SourceStateBlocked)
    {
      if (cres->printLoadMessages)
      {
	fprintf (stderr, "Source StartBlocked %s\n", c->wr->url);
      }
      if (StartSource(cres, c) == 0) break;
      else c = (ChimeraSource)GListGetHead(cres->sources);
    }
    else c = (ChimeraSource)GListGetNext(cres->sources);
  }

  return;
}

/*
 * SourceStopTask
 */
static void
SourceStopTask(closure)
void *closure;
{
  ChimeraSink wp = (ChimeraSink)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  (wp->dhooks.stop)(wp->closure, "");
  wp->wt = NULL;
  
  wp->locked = false;

  if (wp->destroy) SinkDestroy(wp);

  return;
}

/*
 * SourceCreateTask
 */
static void
SourceCreateTask(closure)
void *closure;
{
  ChimeraSink wp = (ChimeraSink)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  if (RequestMatchContent2(wp->ws->wr,
			  GetSourceInfo(wp->ws, "content-type")))
  {
    wp->dstate = DataStateLoading;
    (wp->dhooks.init)(wp, wp->closure);
    
    if (wp->ws->sstate == SourceStateEnded)
    {
      wp->wt = TaskSchedule(wp->ws->cres, SourceEndTask, wp);
    }
  }
  else wp->wt = TaskSchedule(wp->ws->cres, SourceStopTask, wp);
  
  wp->locked = false;

  if (wp->destroy) SinkDestroy(wp);

  return;
}

/*
 * SourceEndTask
 */
static void
SourceEndTask(closure)
void *closure;
{
  ChimeraSink wp = (ChimeraSink)closure;

  wp->wt = NULL;
  if (wp->stop_callbacks) return;

  wp->locked = true;
  
  if (wp->dstate == DataStateLoading)
  {
    wp->dstate = DataStateEnded;
    (wp->dhooks.end)(wp->closure);
  }
  
  wp->locked = false;

  if (wp->destroy) SinkDestroy(wp);

  return;
}

/*
 * TryMemoryCache
 *
 * Rummage around in the list of current downloads to see if what we need
 * is already there.
 */
static ChimeraSource
TryMemoryCache(cres, wr)
ChimeraResources cres;
ChimeraRequest *wr;
{
  ChimeraSource ws;
  char *content;

  /*
   * Input data can make the output different everytime don't even
   * try to find something that has been cached.
   */
  if (wr->input_data != NULL) return(NULL);

  for (ws = (ChimeraSource)GListGetHead(cres->sources); ws != NULL;
       ws = (ChimeraSource)GListGetNext(cres->sources))
  {
    /*
     * Obviously the URL has to match.
     */
    if (RequestCompareURL(ws->wr, wr))
    {
      /*
       * Check to make sure that it hasn't been cancelled and that its
       * reusable.
       */
      if (ws->sstate != SourceStateLame && (ws->cachable || ws->reusable))
      {
	/*
	 * Try to get the content from the download.  If the content-type is
	 * unknown then see if the accept list is the same which is
	 * acceptable (zing!).  If the content-type is known then see if it
	 * is in the accept list for the request.
	 */
	if ((content = GetSourceInfo(ws, "content-type")) == NULL)
	{
	  if (RequestCompareAccept(ws->wr, wr)) return(ws);
	  if (cres->printLoadMessages)
	  {
	    fprintf (stderr, "Data AcceptListDifferent %s\n", wr->url);
	  }
	}
	else if (RequestMatchContent2(ws->wr, content)) return(ws);
	else if (cres->printLoadMessages)
	{
	  fprintf (stderr, "Data ContentDifferent %s\n", wr->url);
	}
      }
      else if (cres->printLoadMessages)
      {
	fprintf (stderr, "Data StateBad %s\n", wr->url);
      }
    }
  }
  
  return(NULL);
}

/*
 * SinkCreate
 *
 * Load some data.  'wr' is eaten by this function.
 */
ChimeraSink
SinkCreate(cres, wr, hooks, closure)
ChimeraResources cres;
ChimeraRequest *wr;
ChimeraSinkHooks *hooks;
void *closure;
{
  ChimeraSource ws = NULL;
  ChimeraSink wp;
  MemPool mp;

  myassert(hooks->init != NULL, "Init function is NULL.");
  myassert(hooks->add != NULL, "Add function is NULL.");
  myassert(hooks->end != NULL, "End function is NULL.");
  myassert(hooks->stop != NULL, "Stop function is NULL.");
  myassert(hooks->message != NULL, "Message function is NULL.");

  if ((ws = TryMemoryCache(cres, wr)) == NULL)
  {
    if ((ws = SourceCreate(cres, wr)) == NULL) return(NULL);
  }
  else
  {
    /*
     * This can be thrown out since the 'ws' obtained from TryMemoryCache
     * contains an identical copy.
     */
    RequestDestroy(wr);
    if (cres->printLoadMessages)
    {
      fprintf (stderr, "Data MemCached %s\n", wr->url);
    }
  }

  mp = MPCreate();
  wp = (ChimeraSink)MPCGet(mp, sizeof(struct ChimeraSinkP));
  wp->mp = mp;
  memcpy(&(wp->dhooks), hooks, sizeof(ChimeraSinkHooks));
  wp->closure = closure;
  wp->dstate = DataStateWaiting;
  wp->ws = ws;

  GListAddTail(ws->datas, wp);
  
  if (ws->sstate == SourceStateLoading || ws->sstate == SourceStateEnded)
  {
    wp->wt = TaskSchedule(ws->cres, SourceCreateTask, wp);
  }

  return(wp);
}

/*
 * SinkDestroy
 */
void
SinkDestroy(wp)
ChimeraSink wp;
{
  ChimeraSource ws = wp->ws;
  ChimeraSink c;

  if (wp->locked)
  {
    wp->destroy = true;
    wp->stop_callbacks = true;
    return;
  }

  if (ws->cres->printLoadMessages)
  {
    fprintf (stderr, "Data Unload %s\n", ws->wr->url);
  }

  wp->dstate = DataStateUnloaded;
  for (c = (ChimeraSink)GListGetHead(ws->datas); c != NULL;
       c = (ChimeraSink)GListGetNext(ws->datas))
  {
    if (c->wt != NULL)
    {
      TaskRemove(ws->cres, c->wt);
      c->wt = NULL;
    }
    if (c->dstate != DataStateUnloaded) break;
  }
  if (c == NULL) SourceDestroy(ws->cres, ws);

  return;
}

/*
 * SinkWasReloaded
 */
bool
SinkWasReloaded(wp)
ChimeraSink wp;
{
  return(wp->ws->wr->reload);
}

/*
 * SourceStop
 */
void
SourceStop(ws, reason)
ChimeraSource ws;
char *reason;
{
  ChimeraSink c;
  GList list;

  ws->sstate = SourceStateLame;

  list = ws->datas;
  for (c = (ChimeraSink)GListGetHead(list); c != NULL;
       c = (ChimeraSink)GListGetNext(list))
  {
    if (!c->stop_callbacks) (c->dhooks.stop)(c->closure, reason);
  }

  StartBlockedSource(ws->cres);

  return;
}

/*
 * SourceEnd
 */
void
SourceEnd(ws)
ChimeraSource ws;
{
  GList list;
  ChimeraSink c;

  myassert(ws->sstate == SourceStateLoading, "End source in wrong state");

  ws->sstate = SourceStateEnded;

  list = ws->datas;
  for (c = (ChimeraSink)GListGetHead(list); c != NULL;
       c = (ChimeraSink)GListGetNext(list))
  {
    if (c->dstate == DataStateLoading && !c->stop_callbacks)
    {
      c->dstate = DataStateEnded;
      (c->dhooks.end)(c->closure);
    }
  }

  if (ws->cres->cc && ws->cachable) CacheWrite(ws->cres->cc, ws);

  StartBlockedSource(ws->cres);

  return;
}

/*
 * SourceAdd
 */
void
SourceAdd(ws)
ChimeraSource ws;
{
  GList list;
  ChimeraSink c;

  myassert(ws->sstate == SourceStateLoading, "Add source in wrong state");

  list = ws->datas;
  for (c = (ChimeraSink)GListGetHead(list); c != NULL;
       c = (ChimeraSink)GListGetNext(list))
  {
    if (c->dstate == DataStateLoading && !c->stop_callbacks)
    {
      (c->dhooks.add)(c->closure);
    }
  }

  return;
}

/*
 * SourceInit
 */
void
SourceInit(ws, usecache)
ChimeraSource ws;
bool usecache;
{
  GList list;
  ChimeraSink c;
  char *url;

  myassert(ws->sstate == SourceStateWaiting, "Init source in wrong state.");
  ws->sstate = SourceStateLoading;
  ws->cachable = usecache;

  url = GetSourceInfo(ws, "x-url");
  myassert(url != NULL, "URL is NULL");
  myassert(GetSourceInfo(ws, "content-type") != NULL,
	   "Content-type is NULL");

  list = ws->datas;
  for (c = (ChimeraSink)GListGetHead(list); c != NULL;
       c = (ChimeraSink)GListGetNext(list))
  {
    c->wt = TaskSchedule(ws->cres, SourceCreateTask, c);
  }

  return;
}

/*
 * SourceSendMessage
 */
void
SourceSendMessage(ws, message)
ChimeraSource ws;
char *message;
{
  GList list;
  ChimeraSink c;

  list = ws->datas;
  for (c = (ChimeraSink)GListGetHead(list); c != NULL;
       c = (ChimeraSink)GListGetNext(list))
  {
    if (c->dhooks.message != NULL && !c->stop_callbacks)
    {
      (c->dhooks.message)(c->closure, message);
    }
  }

  return;
}

/*
 * SinkToResources
 */
ChimeraResources
SinkToResources(wp)
ChimeraSink wp;
{
  return(wp->ws->cres);
}

/*
 * SourceToResources
 */
ChimeraResources
SourceToResources(ws)
ChimeraSource ws;
{
  return(ws->cres);
}

/*
 * SinkGetData
 */
void
SinkGetData(wp, data, len, list)
ChimeraSink wp;
byte **data;
size_t *len;
GList *list;
{
  if (wp->ws->shooks.getdata != NULL)
  {
    (wp->ws->shooks.getdata)(wp->ws->closure, data, len, list);
  }
  return;
}

/*
 * SinkGetInfo
 */
char *
SinkGetInfo(wp, name)
ChimeraSink wp;
char *name;
{
  byte *data;
  size_t len;
  GList list;

  SinkGetData(wp, &data, &len, &list);
  if (list != NULL) return(MIMEGetListField(list, name));
  return(NULL);
}

/*
 * SourceGetData
 */
void
SourceGetData(ws, data, len, list)
ChimeraSource ws;
byte **data;
size_t *len;
GList *list;
{
  if (ws->shooks.getdata != NULL)
  {
    (ws->shooks.getdata)(ws->closure, data, len, list);
  }
  return;
}

/*
 * GetSourceInfo
 */
static char *
GetSourceInfo(ws, name)
ChimeraSource ws;
char *name;
{
  byte *data;
  size_t len;
  GList list;

  SourceGetData(ws, &data, &len, &list);
  if (list == NULL) return(NULL);
  return(MIMEGetListField(list, name));
}

/*
 * SinkCancel
 */
void
SinkCancel(wp, just_stop_callbacks)
ChimeraSink wp;
bool just_stop_callbacks;
{
  ChimeraSource ws = wp->ws;

  wp->stop_callbacks = true;
  if (wp->wt != NULL)
  {
    TaskRemove(ws->cres, wp->wt);
    wp->wt = NULL;
  }

  /*
   * This should probably be more sophisticated.
   */
  if (!just_stop_callbacks)
  {
    ws->sstate = SourceStateLame;
    if (ws->shooks.stop != NULL)
    {
      (ws->shooks.stop)(ws->closure);
    }
  }

  return;
}
