/*
 * mime.c
 *
 * Copyright (c) 1997, John Kilburg <john@cs.unlv.edu>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "port_before.h"

#include <stdio.h>

#ifdef HAVE_STDLIB_H
#include <stdlib.h>
#endif

#include "port_after.h"

#include "common.h"
#include "mime.h"

/*
 * Pseudo-MIME parsing thing.  Hey, it works most of the time, I think.
 *
 * Its not really for MIME.  Its for parsing things that look like
 * what is described in RFC822.
 */

struct MIMEContextP
{
  GList list;
  MemPool mp;
  size_t i;
  size_t doff;
  size_t len;
  bool header_done;
  int http_status;
};

static void parse_header _ArgProto((MIMEContext, byte *, size_t));

MIMEContext
MIMECreateContext()
{
  MIMEContext mc;
  MemPool mp;

  mp = MPCreate();
  mc = (MIMEContext)MPCGet(mp, sizeof(struct MIMEContextP));
  mc->mp = mp;
  mc->list = GListCreateX(mp);
  mc->header_done = false;
  mc->doff = 0;

  return(mc);
}

static void
parse_header(mc, data, len)
MIMEContext mc;
byte *data;
size_t len;
{
  size_t j, k, sep;
  char *cp, *pcp, *scp, *ecp;
  char *line;
  MIMEField *field;

  if (mc->i >= len) return;

  sep = 0;
  cp = NULL;
  scp = (char *)(data + mc->i);

  for (j = 0; j < len; j++)
  {
    pcp = cp;
    cp = scp + j;

/*
    Should check for these at the beginning of the next line to
    extend the value of the previous line.
    if (*cp == ' ') ;
    else if (*cp == '\t') ;
 */
    if (*cp == '\n')
    {
      if (j == 0) k = 0;
      else if (*pcp == '\r') k = j - 1;
      else k = j;

      if (k == 0)
      {
	mc->header_done = true;
	mc->doff = mc->i + j + 1;
      }
      else
      {
	line = (char *)MPGet(mc->mp, k + 1);
	memcpy(line, scp, k);
	line[k] = '\0';

	if (mc->i == 0 && len > 4 && strncmp("HTTP", line, 4) == 0)
	{
	  ecp = line;
	  while (*ecp != '\0' && !isspace8(*ecp))
	  {
	    ecp++;
	  }
	  mc->http_status = atoi(ecp);
	}
	else
	{
	  field = (MIMEField *)MPGet(mc->mp, sizeof(MIMEField));
	  field->name = line;
	  if (sep > 0)
	  {
	    line[sep] = '\0';
	    field->value = line + sep + 1;
	    while (isspace8(*(field->value)))
	    {
	      field->value++;
	    }
	  }
	  else field->value = "";
	  
	  GListAddTail(mc->list, field);
	}
	mc->i += j + 1;
	
	parse_header(mc, data, len);
      }
      
      return;
    }
    else if (*cp == ':' && sep == 0) sep = j;
  }

  return;
}

void
MIMEAddData(mc, data, len)
MIMEContext mc;
byte *data;
size_t len;
{
  if (mc->i >= len) return;
  if (!mc->header_done) parse_header(mc, data, len);
  else mc->i = len;

  return;
}

void
MIMEEndData(mc, data, len)
MIMEContext mc;
byte *data;
size_t len;
{
  MIMEAddData(mc, data, len);
  return;
}

void
MIMEDestroyContext(mc)
MIMEContext mc;
{
  MPDestroy(mc->mp);
  return;
}

/*
 * MIMEGetDataOffset
 */
int
MIMEGetDataOffset(mc, off)
MIMEContext mc;
size_t *off;
{
  if (!mc->header_done) return(-1);
  *off = mc->doff;
  return(0);
}

/*
 * MIMEGetField
 */
char *
MIMEGetField(mc, name)
MIMEContext mc;
char *name;
{
  if (!mc->header_done) return(NULL);
  return(MIMEGetListField(mc->list, name));
}

/*
 * MIMEAddField
 */
void
MIMEAddField(mc, name, value)
MIMEContext mc;
char *name; 
char *value;
{ 
  MIMEAddListField(mc->mp, mc->list, name, value);
  return;
}   

/*
 * ParseExpiresDate
 *
 * This functions sucks.
 */
/*
time_t
ParseExpiresDate(s)
char *s;
{
#ifdef HAVE_MKTIME
  int i;
  time_t expires;
  struct tm tm;
  char buffer[BUFSIZ];
  char month[4], zone[4];
  static char *ms[] =
  { "Jan", "Feb", "Mar", "Apr", "May", "Jun", "Jul", "Aug", "Sep", "Oct",
    "Nov", "Dec" };

  memset(&tm, 0, sizeof(tm));
  if (sscanf(s, "%s %d %s %d %d:%d:%d %s",
	     buffer, &tm.tm_mday, month, &tm.tm_year,
	     &tm.tm_hour, &tm.tm_min, &tm.tm_sec, zone) == 8)
  {
    if (tm.tm_year > 1900) tm.tm_year -= 1900;
    for (i = 0; i < sizeof(ms) / sizeof(ms[0]); i++)
    {
      if (strcmp(month, ms[i]) == 0)
      {
	expires = mktime(&tm);
	break;
      }
    }
  }
  
  return(expires);
#else
  return(0);
#endif
}
*/

/*
 * MIMEGetFieldList
 */
GList
MIMEGetFieldList(mc)
MIMEContext mc;
{
  return(mc->list);
}

/*
 * MIMEGetListField
 */
char *
MIMEGetListField(list, name)
GList list;
char *name;
{
  MIMEField *f;

  for (f = (MIMEField *)GListGetHead(list); f != NULL;
       f = (MIMEField *)GListGetNext(list))
  {
    if (strlen(f->name) == strlen(name) &&
	strcasecmp(f->name, name) == 0)
    {
      return(f->value);
    }
  }

  return(NULL);
}

/*
 * MIMEAddListField
 */
void
MIMEAddListField(mp, list, name, value)
MemPool mp;
GList list;
char *name;
char *value;
{
  MIMEField *f;
  
  f = (MIMEField *)MPGet(mp, sizeof(MIMEField));
  f->name = name;
  f->value = value;
  GListAddHead(list, f);
    
  return;
}

/*
 * MIMEGetHTTPStatus
 */
int
MIMEGetHTTPStatus(mc)
MIMEContext mc;
{
  if (!mc->header_done) return(-1);
  return(mc->http_status);
}
